#include "CHCPlusPlusTraverser.h"

namespace CHCDemo
{

CHCPlusPlusTraverser::CHCPlusPlusTraverser() 
{
}


void CHCPlusPlusTraverser::Traverse()
{
	//-- PART 1: process finished occlusion queries
	while (!mDistanceQueue.empty() || !mQueryQueue.empty())
	{
		while (!mQueryQueue.empty() && 
			   (mQueryQueue.front()->ResultAvailable() || mDistanceQueue.empty()))
		{
			OcclusionQuery *query = mQueryQueue.front();
			mQueryQueue.pop();
			
			// wait until result available
			int visiblePixels = query->GetQueryResult();

			if (visiblePixels > mVisibilityThreshold)
			{
				BvhNode *node = query->GetFrontNode();

				mBvh->MakeParentsVisible(node);
				TraverseNode(node);
			}
			else
			{
				++ mStats.mNumQueryCulledNodes;
			}
		}	

		//-- PART 2: hierarchical traversal
		if (!mDistanceQueue.empty())
		{
			BvhNode *node = mDistanceQueue.top();
			mDistanceQueue.pop();
	
			if (mBvh->IsWithinViewFrustum(node))
			{
				// for near plane intersecting bounding box possible 
				// wrong results => skip occlusion query
				if (IntersectsNearPlane(node))
				{
					// update node's visited flag
					node->SetLastVisitedFrame(mFrameId);
					node->SetVisible(true);
					mBvh->MakeParentsVisible(node);

					TraverseNode(node);
				}
				else
				{		
					// identify previously visible nodes
					const bool wasVisible = node->IsVisible() && (node->GetLastVisitedFrame() == mFrameId - 1);
					
					// identify nodes that we cannot skip queries for
					const bool leafOrWasInvisible = 
						(node->GetLastVisitedFrame() != mFrameId) && (!wasVisible || node->IsVirtualLeaf());

					// reset node's visibility classification 
					node->SetVisible(false);

					// update node's visited flag
					node->SetLastVisitedFrame(mFrameId);
					
					// skip testing previously visible interior nodes
					if (leafOrWasInvisible)
					{
						OcclusionQuery *query = IssueOcclusionQuery(node, wasVisible);
						mQueryQueue.push(query);
					}
					
					// always traverse a node if it was visible
					if (wasVisible)
						TraverseNode(node);
				}
			}
			else
			{
				// for stats
				++ mStats.mNumFrustumCulledNodes;
			}
		}
	}
}

}